theory NikBoolExpr
(*
 *  Niklaus boolean expressions (conditions for if and while)
 *)
imports NikExpressions

begin

(* 
 *  A boolean expression may consist in the comparison 
 *  of two arithmetic expressions or in a boolean value
 *)
datatype boolexp = 
    Equals expression expression       (infix ".=." 55)
  | Differs expression expression      (infix ".\<noteq>." 55)
  | LowerThan expression expression    (infix ".<." 55)
  | LowerEqual expression expression   (infix ".\<le>." 55)
  | GreaterThan expression expression  (infix ".>." 55)
  | GreaterEqual expression expression (infix ".\<ge>." 55)
  | Boolean bool                       ("B")

(* Semantics of boolean expressions is mapped to the usual operators on integers *)
fun evalbool :: "boolexp \<Rightarrow> environment \<Rightarrow> bool"
where
    "evalbool (e1 .=. e2) env =  (evaluate e1 env = evaluate e2 env)"
 |  "evalbool (e1 .\<noteq>. e2) env =  (evaluate e1 env \<noteq> evaluate e2 env)"
 |  "evalbool (e1 .<. e2) env =  (evaluate e1 env < evaluate e2 env)"
 |  "evalbool (e1 .\<le>. e2) env =  (evaluate e1 env \<le> evaluate e2 env)"
 |  "evalbool (e1 .>. e2) env =  (evaluate e1 env > evaluate e2 env)"
 |  "evalbool (e1 .\<ge>. e2) env =  (evaluate e1 env \<ge> evaluate e2 env)"
 |  "evalbool (B b) env = b"

(* Checks on some expressions *)
value "evalbool (N 2 .+. V ''x'' .=. N 4) ((\<lambda>x.0)(''x'':=5))"
value "evalbool (N 2 .+. V ''x'' .=. N 4)  ((\<lambda>x.0)(''x'':=2))"
value "evalbool (B True) (\<lambda>x. 0)"

end

