theory NikDenBig

imports NikDenotational NikBigStep

begin

(* Equivalence between while loop and unrolled while loop *)
lemma unroll_while: "\<lbrakk> while c do body done \<rbrakk> = \<lbrakk> if c then body;; while c do body done else Nop fi \<rbrakk>"
proof -
  let ?w = "while c do body done"
  let ?wf = "while_func (evalbool c) \<lbrakk> body \<rbrakk>"
  from While_den have "\<lbrakk> ?w \<rbrakk> = lfp ?wf" .
  also have "... = ?wf (lfp ?wf)" by (rule lfp_unfold[OF while_mono])
  also have "... = ?wf \<lbrakk> ?w \<rbrakk>" by simp
  also have "... = {(pre, post). if (evalbool c) pre
                                 then (pre,post) \<in> \<lbrakk> body \<rbrakk> O \<lbrakk> ?w \<rbrakk>
                                 else post = pre }" unfolding while_func_def ..
  also have "... = {(pre, post). if (evalbool c) pre
                                 then (pre, post) \<in> \<lbrakk> body ;; ?w \<rbrakk>
                                 else post = pre }" using Seq_den by simp
  also have "... = \<lbrakk> if c then body;; ?w else Nop fi \<rbrakk>" by auto
  finally show ?thesis .
qed

lemma big_step_in_den: "(prog, pre) \<leadsto> post \<Longrightarrow> (pre, post) \<in> \<lbrakk> prog \<rbrakk>"
proof (induction rule: big_step_induct) print_cases
  case WhileFalse thus ?case using unroll_while by simp
next
  case WhileTrue print_facts
  show ?case unfolding unroll_while using WhileTrue by auto
qed auto

(* Make big step semantics a relation *)
abbreviation big_step_rel:: "instruction \<Rightarrow> instr_den"
where "big_step_rel instr \<equiv> {(pre, post). (instr, pre) \<leadsto> post}"


lemma den_in_big_step: "(pre, post) \<in> \<lbrakk> instr \<rbrakk> \<Longrightarrow> (pre, post) \<in> big_step_rel instr"
proof (induction instr arbitrary: pre post) print_cases
  case Nop thus ?case by auto
next
  case (VarDecl v) thus ?case (* by force *)
  proof -
    from VarDecl and nik_den.simps(2) have "post = pre(v:=0)" by auto
    hence "(pre, post) \<in> big_step_rel (var v)" by auto
    thus ?thesis .
  qed
next
  case (Assign v val) thus ?case
  proof -
    from Assign and nik_den.simps(3) have "post = pre(v:=evaluate val pre)" by auto
    hence "(pre, post) \<in> big_step_rel (v <- val)" by auto
    thus ?thesis .
  qed
next
  case (Alternative cond iT iF) thus ?case
  apply (unfold nik_den.simps(5))
  apply simp
  apply(case_tac "evalbool cond pre")
  apply auto
  done
next
  case (Seq i1 i2) thus ?case (* by force *) print_facts
  proof -
    from Seq.prems obtain p where
        "(pre, p) \<in> \<lbrakk> i1 \<rbrakk>"
    and "(p, post) \<in> \<lbrakk> i2 \<rbrakk>" by auto
    hence "(pre, p) \<in> big_step_rel i1" 
    and "(p, post) \<in> big_step_rel i2"
    using Seq.IH by auto
    hence "(pre, post) \<in> big_step_rel (i1;; i2)" by auto
    thus ?thesis .
  qed
next
  case (While cond body) print_facts
    let ?bigw = "big_step_rel (while cond do body done)"
    let ?wf = "while_func (evalbool cond) \<lbrakk> body \<rbrakk>"
    have "?wf ?bigw \<subseteq> ?bigw" unfolding while_func_def using While.IH by auto
    from lfp_lowerbound[of ?wf, OF this] and While.prems show ?case by auto
qed

theorem den_is_big_step: "((instr, pre) \<leadsto> post) = ((pre, post) \<in> \<lbrakk> instr \<rbrakk>)"
(* by (auto simp add: big_step_in_den den_in_big_step[simplified]) *)
proof -
  have "(pre, post) \<in> big_step_rel instr \<Longrightarrow> (instr, pre) \<leadsto> post" by simp
  moreover from den_in_big_step have "(pre, post) \<in> \<lbrakk> instr \<rbrakk> \<Longrightarrow> (pre, post) \<in> big_step_rel instr " .
  ultimately have 1: "(pre, post) \<in> \<lbrakk> instr \<rbrakk> \<Longrightarrow> (instr, pre) \<leadsto> post" .
  from big_step_in_den have 2: "(instr, pre) \<leadsto> post \<Longrightarrow> (pre, post) \<in> \<lbrakk> instr \<rbrakk>" .
  from 1 and 2 show ?thesis by auto
qed

end

