theory NikExprSimplify
(*
 * Simplification of Niklaus arithmetic expressions
 * This is a template to be completed by the students
 *)
imports NikExpressions

begin

(*
 * Simplifying an expression amounts to replacing:
 *  - all constant expressions by their value (e.g. N 2 .+. N 3 by N 5)
 *  - N 0 .+. x and x .+. N 0 by x
 *  - N 0 .*. x and x .*. N 0 by N 0
 *  - N 1 .*. x and x .*. N 1 by x
 *  - N 0 .\<div>. x by N 0
 *  - x .\<div>. N 1 by x
 *)
fun simplify:: "expression \<Rightarrow> expression"
where
    (* A binary operator can be simplified if both arguments simplify to numbers *)
    "simplify (a .+. b) = (
       let (u', v') = (simplify a, simplify b) in
       if u' = (N 0) then v'
       else if v' = (N 0) then u'
       else
       case (u', v') of
           (N va, N vb) \<Rightarrow> N (va + vb)  (* arguments are numbers \<Rightarrow> compute the result *)
         | (sa, sb) \<Rightarrow> sa .+. sb        (* else return the sum of the simplified arguments *)
    )"
  | (* Other cases ... *)
  (* Other expressions (variables and constants) cannot be simplified and are left as is *)
  | "simplify expr = expr"

(* Some checks *)
value "simplify ((N 2) .*. (N 3))"
value "simplify (N 2 .*. N 1 .+. V ''x'')"
value "simplify (N 0 .*. (N 1 .+. V ''x''))"
value "simplify ((N 2 .-. N 2) .*. (N 1 .+. V ''x''))"

(* Proof that simplify does not change the semantics of expressions *)
theorem "evaluate (simplify expr) env = evaluate expr env"
(* by (induction expr, auto split:expression.split) *)
sorry

(* A predicate telling whether an expression is optimal, i.e. does not
   contain any operator applied to constants. *)
fun optimal_expr :: "expression \<Rightarrow> bool"
where
    "optimal_expr (N a) = True"
  | "optimal_expr (V x) = True"
  | (* Other cases... *)

(* Check that optimal_expr behaves as expected *)
value "optimal_expr (N 3)"
value "optimal_expr (N 3 .+. V ''x'')"
value "optimal_expr (N 3 .+. N 5)"
value "optimal_expr ((N 3 .+. V ''x'') .*. V ''y'')"
value "optimal_expr ((N 3 .+. N 2) .*. V ''y'')"

(* Prove that simplify yields optimal expressions *)
theorem "optimal_expr (simplify expr)"
sorry

end

