theory NikInstructions
(*
 *  Abstract syntax of the instructions in Niklaus
 *)
imports NikExpressions NikBoolExpr

begin

(*
 * Without the Read and Write instructions, the state 
 * of the Niklaus execution machine is reduced to the
 * environment (valuation of the variables)
 *)
type_synonym state = "environment"

(*
 *  Instructions are:
 *    - Nop, does nothing, useful for if _ then _ else when a branch is empty
 *    - VarDecl v (abbreviated into "var v"), declares a variable (initial value is 0)
 *    - Seq i1 i2 (abbreviated as "i1;; i2) executes i1 and i2 in sequence
 *    - Alternative c i1 i2 ("if c then i1 else i2 fi"), executes i1 if c is true, else executes i2
 *    - While c body ("while c do body done"), executes body if c is true, then executes itself again
 *)
datatype instruction = 
    Nop
  | VarDecl varname                ("var")
  | Assign varname expression      ("_ <- _" [1000, 65] 65)
  | Seq instruction instruction    ("_ ;; _" [65, 60] 60)
  | Alternative boolexp instruction instruction
                                   ("if _ then _ else _ fi")
  | While boolexp instruction      ("while _ do _ done")

(* Check that we can write some simple instructions with this syntax *)
term "var ''x''"
term "if V ''x'' .<. N 2 then ''x'' <- N 2 else ''x''<- V ''x'' .-. N 1 fi"

end

