theory NikPGCD

imports NikAxiomatic PGCD

begin
(* Proof that Euclid's algorithm programmed in Niklaus computes the GCD of two integers *)

(* First, the if_then_else in the loop *)
abbreviation if_x_y:: "instruction"
where "if_x_y \<equiv> 
      if V ''x'' .<. V ''y'' then
        ''y'' <- V ''y'' .-. V ''x''
      else
        ''x'' <- V ''x'' .-. V ''y''
      fi"

(* Prove that this if_then_else preserves gcd *)
lemma gcd_if:
  "\<turnstile> {\<lambda>s. is_gcd (s ''x'') (s ''y'') k}
      if_x_y
     {\<lambda>s. is_gcd (s ''x'') (s ''y'') k}"
apply (rule If)
apply (rule Assign')
apply simp
apply (insert is_gcd_b_a)
apply simp

apply (rule Assign')
apply simp
apply (insert is_gcd_a_b)
apply simp
done

(* The while loop *)
abbreviation while_x_y:: "instruction"
where "while_x_y \<equiv> while V ''x'' .\<noteq>. V ''y'' do if_x_y done"

(* Prove that this loop preserves gcd *)
lemma gcd_while:
  "\<turnstile> {\<lambda>s. is_gcd (s ''x'') (s ''y'') k}
       while_x_y
     {\<lambda>s. is_gcd (s ''x'') (s ''y'') k \<and> (s ''x'') = (s ''y'')}"
apply (rule While')
apply simp
apply (rule strengthen_pre[OF _ gcd_if])
apply simp
apply simp
apply auto
done

(* Useful lemma for concluding *)
lemma final_gcd: "is_gcd x y k \<and> x = y \<Longrightarrow> is_gcd x y x"
by (insert is_gcd_a_a_a, auto)

(* Prove that at the end of the loop, x is the gcd of x and y *)
lemma gcd_while2:
  "\<turnstile> {\<lambda>s. is_gcd (s ''x'') (s ''y'') k}
       while_x_y
     {\<lambda>s. is_gcd (s ''x'') (s ''y'') (s ''x'')}"
apply (rule weaken_post[OF gcd_while])
apply (insert final_gcd)
apply auto
done

end
