theory PGCD

imports Main

begin

definition divides:: "int \<Rightarrow> int \<Rightarrow> bool"
where "divides a b = (\<exists>k::int. b = k * a)"

definition common_div :: "int => int => int => bool"
where "common_div a b k \<equiv> divides k a \<and> divides k b"

definition is_gcd :: "int => int => int => bool" 
where "is_gcd a b g \<equiv> common_div a b g 
                      \<and> (\<forall>p::int. common_div a b p \<longrightarrow> divides p g)"

theorem is_gcd_a_a_a: "is_gcd a a a"
by (simp add: common_div_def divides_def is_gcd_def)

(* More detailed proof *)
theorem "is_gcd a a a"
proof -
  from divides_def have "divides a a" by simp
  hence 1:"common_div a a a" using common_div_def by simp
  {fix p assume "common_div a a p"
   hence "divides p a" using common_div_def by simp
  }
  with 1 show ?thesis using is_gcd_def by simp
qed

theorem common_div_a_b:
  "common_div a b k \<longrightarrow> common_div (a-b) b k"
by (metis common_div_def divides_def int_distrib(3))

(* More detailed proof *)
theorem "common_div a b k \<longrightarrow> common_div (a-b) b k"
proof
  assume "common_div a b k"
  hence 1: "divides k a \<and> divides k b" using common_div_def by simp
  hence "(\<exists>pa::int. a = pa*k) \<and> (\<exists>pb::int. b = pb*k)" using divides_def by simp
  then obtain pa pb where "a = pa*k" and "b = pb*k" by auto
  hence "(a-b) = (pa-pb) * k" by (simp add: int_distrib)
  hence "divides k (a-b)" using divides_def by simp
  thus "common_div (a-b) b k" using 1 and common_div_def by simp
qed

theorem common_div_commutes:
  "common_div a b k = common_div b a k"
using common_div_def by auto

(* Other proof *)
theorem "common_div a b k = common_div b a k"
proof
  assume "common_div a b k"
  from this and common_div_def show "common_div b a k" by simp
next
  assume "common_div b a k"
  from this and common_div_def show "common_div a b k" by simp
qed

theorem is_gcd_commutes:
  "is_gcd a b k = is_gcd b a k"
by (simp add: common_div_commutes is_gcd_def)

(* Other proof *)
theorem "is_gcd a b k = is_gcd b a k"
proof
  assume "is_gcd a b k"
  then show "is_gcd b a k" using is_gcd_def and common_div_commutes by simp
next
  assume "is_gcd b a k"
  then show "is_gcd a b k" using is_gcd_def and common_div_commutes by simp
qed

theorem is_gcd_a_b:
  "is_gcd a b k \<longrightarrow> is_gcd (a-b) b k"
proof
  assume h: "is_gcd a b k"
  hence "common_div a b k" using is_gcd_def by simp
  hence 1: "common_div (a-b) b k" using common_div_a_b by simp
  (* Show that any other common divisor of (a-b) and b divides k *)
  { fix p
	  assume h': "common_div (a-b) b p"
	  hence "(\<exists>pab::int. (a-b) = pab * p) \<and> (\<exists>pb::int. b = pb * p)" using common_div_def and divides_def by simp
	  then obtain pab pb where "(a-b) = pab * p" and "b = pb * p" by auto
	  hence "(a-b) + b = pab * p + pb * p" by simp
	  hence "a = pab * p + pb * p" by simp
	  hence "a = (pab+pb) * p" using int_distrib by simp
	  hence "divides p a" using divides_def by simp
	  hence "common_div a b p" using h' and common_div_def by simp
	  hence "divides p k" using h and is_gcd_def by simp
  }
  hence "\<forall>p::int. common_div (a-b) b p \<longrightarrow> divides p k" by simp
  thus "is_gcd (a-b) b k" using 1 and is_gcd_def by simp
qed

theorem is_gcd_b_a:
  "is_gcd a b k \<longrightarrow> is_gcd a (b-a) k"
using is_gcd_a_b is_gcd_commutes by auto

(* More detailed proof *)
theorem "is_gcd a b k \<longrightarrow> is_gcd a (b-a) k"
proof
  assume "is_gcd a b k"
  hence "is_gcd b a k" using is_gcd_commutes by simp
  hence "is_gcd (b-a) a k" using is_gcd_a_b by simp
  thus "is_gcd a (b-a) k" using is_gcd_commutes by simp
qed

end
